#
#  gpsman --- GPS Manager: a manager for GPS receiver data
#
#  Copyright (c) 2000 Miguel Filgueiras (mig@ncc.up.pt) / Universidade do Porto
#
#    This program is free software; you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation; either version 2 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.
#
#  File: trtort.tcl
#  Last change:  2 October 2000
#
# Includes contributions by Brian Baulch (baulchb@onthenet.com.au)
#  marked "BSB contribution"
#

## map tags used, with $w a window path, $n a natural number:
#    lines in converted RT: exp=$w expRT=$w line
#    turnpoints in converted RT: exp=$w expRT=$w sq2 lab=$n.$n
#    TR elements: exp=$w expTR=$w (in addition to normal tags)
###

proc GMTRtoRT {window} {
    # create dialog to control conversion from track to route
    #  $window is the track edit/show window
    global DPOSX DPOSY COLOUR GMTRTPs GMTRDatum GMTRData TRTPoints TRDatum \
	    TRDispl TXT MESS Map EdWindow MAPSCLENGTH TRCvNPoints TRCvNPsOld\
	    TRCvRTDispl TRCvTRDispl TRCvRT TRCvTPs TRCvDatum \
	    MouseOnTRCvScale

    if { [winfo exists $EdWindow(RT)] } { Raise $EdWindow(RT) ; bell ; return }
    set w ${window}.trtort
    if { [winfo exists $w] } { Raise $w ; bell ; return }

    if {! [string compare $window $EdWindow(TR)] } {
	if {! [string compare [set TRCvTPs($w) $GMTRTPs] ""] } {
	    GMMessage $MESS(voidTR)
	    return
	}
	set mapped [lindex $GMTRData end]
	set TRCvDatum($w) $GMTRDatum
    } else {
	set ixt [IndexNamed TR [$window.fr.fr1.id get]]
	set TRCvTPs($w) $TRTPoints($ixt) ; set TRCvDatum($w) $TRDatum($ixt)
	set mapped $TRDispl($ixt)
    }
    set trname [string trim [$window.fr.fr1.id get]]
    if {! [string compare "$trname" ""] } {
	set trname $TXT(unnamed)
    }
    set TRCvNPoints($w) 2 ; set TRCvNPsOld($w) 1
    set TRCvRT($w) [TRCvRoute $w $TRCvNPoints($w)]
    set TRCvRTDispl($w) 0 ; set TRCvTRDispl($w) $mapped
    set MouseOnTRCvScale 0
    set ntps [llength $TRCvTPs($w)]

    toplevel $w
    wm protocol $w WM_DELETE_WINDOW "TRCvCancel $w"
    wm title $w "GPS Manager: $TXT(TRtoRT)"
    wm transient $w
    set x [expr $DPOSX+100] ; set y [expr $DPOSY+100]
    wm geometry $w +$x+$y

    frame $w.fr -relief flat -borderwidth 5 -bg $COLOUR(dialbg)
    label $w.fr.from -text [format $TXT(fromTR) $trname]
    frame $w.fr.sep -height 6 -bg $COLOUR(dialbg) \
	    -relief flat -borderwidth 0
    label $w.fr.npts -text $TXT(TRRTnpoints)
    scale $w.fr.npscl -orient horizontal -from 2 -to $ntps -width 5 \
	    -length $MAPSCLENGTH -resolution 1 -showvalue 1 \
	    -variable TRCvNPoints($w)
    $w.fr.npscl set 2
    bind $w.fr.npscl <Leave> "TRCvNumberChange $w"
    bind $w.fr.npscl <Button-1> { set MouseOnTRCvScale 1 }
    bind $w.fr.npscl <ButtonRelease-1> { set MouseOnTRCvScale 0 }

    frame $w.fr.act -relief flat -borderwidth 0
    checkbutton $w.fr.act.drt -text $TXT(TRRTdispl) -variable TRCvRTDispl($w) \
	    -onvalue 1 -offvalue 0 -selectcolor $COLOUR(check) \
	    -command "TRCvDisplRT $w"
    checkbutton $w.fr.act.dtr -text $TXT(TRTRdispl) -variable TRCvTRDispl($w) \
	    -onvalue 1 -offvalue 0 -selectcolor $COLOUR(check) \
	    -command "TRCvDisplTR $w"
    # if TR is mapped do not allow it to be unmapped
    if { $mapped } { $w.fr.act.dtr configure -state disabled }

    frame $w.fr.bts -relief flat -borderwidth 0
    button $w.fr.bts.ok -text Ok -command "TRCvCommit $w"
    button $w.fr.bts.cnc -text $TXT(cancel) -command "TRCvCancel $w"

    pack $w.fr.from $w.fr.sep $w.fr.npts $w.fr.npscl -side top
    pack $w.fr.act.drt $w.fr.act.dtr -side left -padx 3
    pack $w.fr.bts.ok $w.fr.bts.cnc -side left -padx 3
    pack $w.fr.act $w.fr.bts -side top -pady 5
    pack $w.fr
    return
}

proc TRCvNumberChange {w} {
    global MouseOnTRCvScale TRCvNPoints TRCvNPsOld TRCvRT

    if { $MouseOnTRCvScale || $TRCvNPoints($w)==$TRCvNPsOld($w) } { return }
    SetCursor ". $w" watch
    set TRCvNPsOld($w) $TRCvNPoints($w)
    set TRCvRT($w) [TRCvRoute $w $TRCvNPoints($w)]
    ResetCursor ". $w"
    TRCvDisplRT $w
    return
}

proc TRCvCommit {w} {
    # define WPs from points in conversion result and open RT edit window
    # fail if RT edit window already exists
    global TRCvRT TRCvDatum EdWindow PositionFormat CREATIONDATE \
	    WPRoute GRName DEFAULTSYMBOL DEFAULTDISPOPT TXT MYGPS

    if { [winfo exists $EdWindow(RT)] } { Raise $EdWindow(RT) ; bell ; return }
    SetCursor ". $w" watch
    set datum $TRCvDatum($w)
    set type [PosType $PositionFormat]
    # BSB contribution: empty date if Lowrance receiver
    if { $CREATIONDATE && [string compare $MYGPS "Lowrance"] } {
	set date [Now]
    } else { set date "" }
    set wps "" ; set n 0
    foreach p $TRCvRT($w) {
	set id [format "ZT%04d" $n]
	while { [IndexNamed WP $id] != -1 } {
	    incr n ; set id [format "ZT%04d" $n]
	}
	incr n
	set data [FormData WP "Name PFrmt Posn Datum Date Alt" \
		     [list $id $PositionFormat \
		       [CreatePos [lindex $p 0] [lindex $p 1] \
		                  $PositionFormat $type $datum] \
		       $datum $date [lindex $p 6]]]
	set ix [CreateItem WP $data]
	set WPRoute($ix) ""
	lappend wps $id
    }
    set grname [format "TR-RT %d" $n]
    while { [IndexNamed GR $grname] != -1 } {
	incr n ; set grname [format "TR-RT %d" $n]
    }
    set data [FormData GR "Name Obs Conts" \
	         [list $grname $TXT(obsTRToRT) [list [list WP $wps]]]]
    CreateItem GR $data
    ResetCursor ". $w"
    TRCvCancel $w
    if { [winfo exists $EdWindow(RT)] } { Raise $EdWindow(RT) ; bell ; return }
    GMRoute -1 {create revert cancel} \
	    [FormData RT "WPoints Displ" [list $wps 0]]
    return
}

proc TRCvCancel {w} {
    # finish off conversion from TR to RT on window $w
    global TRCvNPoints TRCvNPsOld TRCvRTDispl TRCvTRDispl TRCvRT TRCvTPs \
	    TRCvDatum Map

    if { $TRCvRTDispl($w) || $TRCvTRDispl($w) } {
	$Map delete exp=$w
	SetMapBounds
    }
    destroy $w
    unset TRCvNPoints($w) ; unset TRCvRTDispl($w) ; unset TRCvTRDispl($w)
    unset TRCvRT($w) ; unset TRCvTPs($w) ; unset TRCvDatum($w)
    unset TRCvNPsOld($w)
    return
}

proc TRCvDisplRT {w} {
    global Map TRCvRTDispl TRCvRT TRCvDatum

    $Map delete expRT=$w
    if { $TRCvRTDispl($w) } {
	SetCursor ". $w" watch
	set tgspt [list exp=$w expRT=$w sq2]
	set tgsln [list exp=$w expRT=$w line]
	set ps $TRCvRT($w) ; set datum $TRCvDatum($w)
	set p1 [lindex $ps 0]
	set pp1 [MapFromPosn [lindex $p1 0] [lindex $p1 1] $datum]
	set x [lindex $pp1 0] ; set y [lindex $pp1 1]
	$Map create oval [expr $x-2] [expr $y-2] \
		[expr $x+2] [expr $y+2] -fill black -outline black \
		-tags [concat lab=$w.0 $tgspt]
	set i 1
	foreach p2 [lrange $ps 1 end] {
	    set pp2 [MapFromPosn [lindex $p2 0] [lindex $p2 1] $datum]
	    set x2 [lindex $pp2 0] ; set y2 [lindex $pp2 1]
	    $Map create oval [expr $x2-2] [expr $y2-2] \
		[expr $x2+2] [expr $y2+2] -fill black -outline black \
		-tags [concat lab=$w.$i $tgspt]
	    $Map create line $x $y $x2 $y2 -arrow last -smooth 0 \
		    -fill black -width 2 -tags $tgsln
	    set x $x2 ; set y $y2
	    incr i
	}
	ResetCursor ". $w"
    }
    SetMapBounds
    return
}

proc TRCvDisplTR {w} {
    # map/unmap TR during this conversion
    # TR was not mapped when the conversion started
    global Map TRCvTRDispl TRCvTPs TRCvDatum

    if { $TRCvTRDispl($w) } {
	SetCursor $w watch
	PutMapTREls -1 $TRCvTPs($w) $TRCvDatum($w) [list exp=$w expTR=$w]
	ResetCursor $w
    } else {
	$Map delete expTR=$w
    }
    SetMapBounds
    return
}

proc TRCvRoute {w n} {
    # using information from window $w, convert list of "positions" in
    #  given datum to list of $n>=2 "positions"
    # "position" means list starting with lat long in degrees
    # procs SetDatumData and ComputeDistBearFD are called (indirectly)
    global TRCvTPs TRCvDatum

    set ps $TRCvTPs($w)
    if { $n<3 || [llength $ps]<3 } {
	return [list [lindex $ps 0] [lindex $ps end]]
    }
    SetDatumData $TRCvDatum($w)
    set ss [TRCvRouteSegms [SplitPositions $ps] [expr $n-1]]
    set r [list [lindex [lindex [lindex $ss 0] 2] 0]]
    foreach s $ss {
	lappend r [lindex [lindex $s 2] end]
    }
    return $r
}

proc TRCvRouteSegms {ss n} {
    # breaks a route segment into $n>=2 segments
    # a segment is a list of:
    #     maximum distance from a point to line between 1st and last points
    #     index of a point at that distance
    #     "positions" of points

    while { $n != 2 } {
	set mx 0 ; set im 0 ; set i 0
	foreach s $ss {
	    if { [set d [lindex $s 0]] > $mx } {
		set mx $d ; set im $i
	    }
	    incr i
	}
	if { $mx == 0 } { return $ss }
	set s [lindex $ss $im]
	set is [lindex $s 1] ; set ps [lindex $s 2]
	set ss [concat [lrange $ss 0 [expr $im-1]] \
		       [SplitPositionsAt $is $ps] \
		       [lrange $ss [expr $im+1] end]]
	incr n -1
    }
    return $ss
}

proc SplitPositions {ps} {
    # split "positions" into two segments at maximum distance to line between
    #  end points

    set dfp [FurthestPoint $ps]
    set im [lindex $dfp 1]
    return [SplitPositionsAt $im $ps]
}

proc SplitPositionsAt {im ps} {
    # split "positions" into two segments at given index $im

    set ps1 [lrange $ps 0 $im] ; set ps2 [lrange $ps $im end]
    return [list [linsert [FurthestPoint $ps1] end $ps1] \
	         [linsert [FurthestPoint $ps2] end $ps2]]
}

proc FurthestPoint {ps} {
    # find a point that is furthest from the line between 1st and last points
    # return list with maximum distance and index of point

    if { [set l [llength $ps]] < 3 } { return [list 0 0] }
    set mx 0 ; set i 1
    set pt0 [lindex $ps 0] ; set ptn [lindex $ps end]
    foreach pt [lrange $ps 1 [expr $l-2]] {
	set d [DistToLineFD $pt $pt0 $ptn]
	if { $d >= $mx } {
	    set im $i ; set mx $d
	}
	incr i
    }
    if { $mx == 0 } { set im [expr $l/2] }
    return [list $mx $im]
}

proc DistToLineFD {p p1 p2} {
    # find distance from point $p to line between $p1 and $p2
    #  $p, $p1 and $p2 are "positions"

    set db [ComputeDistBearFD $p1 $p]
    set a [expr [lindex [ComputeDistBearFD $p1 $p2] 1]-[lindex $db 1]]
    return [expr abs(sin($a*0.01745329251994329576)*[lindex $db 0])]
}
